from errno import ENOENT
from tempfile import mkdtemp
from unittest import TestCase
import sys

from bp.filepath import FilePath

from mkenv import find, make
from mkenv.common import Locator
from mkenv.tests.utils import CLIMixin


class TestMake(CLIMixin, TestCase):

    cli = make

    def test_make_creates_an_env_with_the_given_name(self):
        self.assertFalse(self.locator.for_name("made").exists)
        self.run_cli(["made"])
        self.assertTrue(self.locator.for_name("made").exists)

    def test_make_t_creates_a_global_temporary_environment_OSError(self):
        temporary = self.locator.temporary()
        self.assertFalse(temporary.exists)

        def real_fake_remove(path):
            """
            os.remove will raise an OSError, not an IOError.

            """

            raise OSError(ENOENT, "NOPE")

        MemoryPath = temporary.path.__class__
        remove, MemoryPath.remove = MemoryPath.remove, real_fake_remove
        self.addCleanup(setattr, MemoryPath, "remove", remove)

        self.run_cli(["--temporary"])
        self.assertTrue(temporary.exists)

    def test_make_t_creates_a_global_temporary_environment_IOError(self):
        """
        If we use os.remove, this one actually should never happen, since that
        will raise OSError, but we ensure we cover both cases.

        """

        temporary = self.locator.temporary()
        self.assertFalse(temporary.exists)

        self.run_cli(["--temporary"])
        self.assertTrue(temporary.exists)

    def test_make_t_recreates_the_environment_if_it_exists(self):
        temporary = self.locator.temporary()
        self.assertFalse(temporary.exists)
        self.run_cli(["--temporary"])
        self.assertTrue(temporary.exists)

        foo = temporary.path.child("foo")
        foo.setContent("testing123")
        self.assertTrue(foo.exists())

        self.run_cli(["--temporary"])
        self.assertTrue(temporary.exists)
        self.assertFalse(foo.exists())

    def test_cannot_specify_both_name_and_temporary(self):
        stdin, stdout, stderr = self.run_cli(
            ["--temporary", "foo"], exit_status=2,
        )
        self.assertTrue(
            stderr.endswith(
                "specify only one of '-t / --temp / --temporary' or 'name'\n"
            ), msg=stderr,
        )

    def test_recreate(self):
        virtualenv = self.locator.for_name("something")
        self.assertFalse(virtualenv.exists)

        virtualenv.create()
        virtualenv.path.child("thing").setContent("")

        self.run_cli(["--recreate", "something"])
        self.assertTrue(virtualenv.exists)
        self.assertFalse(virtualenv.path.child("thing").exists())

    def test_install_and_requirements(self):
        self.run_cli(["-i", "foo", "-i", "bar", "-r", "reqs.txt", "bla"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed.get(self.locator.for_name("bla")),
            [(("foo", "bar"), ("reqs.txt",))],
        )

    def test_install_default_name(self):
        """
        If you install one single package and don't specify a name, the name of
        the installed package is used.

        """

        self.run_cli(["-i", "foo"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed.get(self.locator.for_name("foo")), [(("foo",), ())],
        )

    def test_install_default_name_with_version_specification(self):
        self.run_cli(["-i", "thing[foo]>=2,<3"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed.get(self.locator.for_name("thing")),
            [(("thing[foo]>=2,<3",), ())],
        )


class TestIntegration(TestCase):
    def setUp(self):
        self.root = FilePath(mkdtemp())
        self.addCleanup(self.root.remove)

        # Fucking click.
        stdout = sys.stdout
        self.addCleanup(lambda: setattr(sys, "stdout", stdout))

    def test_it_works(self):
        with self.root.child("make_stdout").open("w") as stdout:
            sys.stdout = stdout

            try:
                make.main(
                    args=[
                        "--root", self.root.path,
                        "mkenv-unittest-should-be-deleted",
                    ],
                )
            except SystemExit:
                pass

        with self.root.child("find_stdout").open("w") as stdout:
            sys.stdout = stdout

            try:
                find.main(
                    [
                        "--root", self.root.path,
                        "--existing-only",
                        "name", "mkenv-unittest-should-be-deleted",
                    ],
                )
            except SystemExit:
                pass

        locator = Locator(root=self.root)
        virtualenv = locator.for_name("mkenv-unittest-should-be-deleted")
        self.assertEqual(
            self.root.child("find_stdout").getContent(),
            virtualenv.path.path + "\n",
        )



from unittest import TestCase

from bp.filepath import FilePath
from bp.memory import MemoryPath

from mkenv import find
from mkenv.tests.utils import CLIMixin


class TestFind(CLIMixin, TestCase):

    cli = find

    def test_find_directory_finds_envs_by_directory(self):
        this_dir = FilePath(__file__).parent()
        stdin, stdout, stderr = self.run_cli(["directory", this_dir.path])
        self.assertEqual(
            (stdin, stdout, stderr),
            ("", self.locator.for_directory(this_dir).path.path + "\n", ""),
        )

    def test_find_directory_defaults_to_cwd(self):
        stdin, stdout, stderr = self.run_cli(["directory"])
        self.assertEqual(
            (stdin, stdout, stderr), (
                "",
                self.locator.for_directory(FilePath(".")).path.path + "\n",
                "",
            ),
        )

    def test_find_name_finds_envs_by_name(self):
        stdin, stdout, stderr = self.run_cli(["name", "bla"])
        self.assertEqual(
            (stdin, stdout, stderr),
            ("", self.locator.for_name("bla").path.path + "\n", ""),
        )

    def test_find_without_args_finds_the_virtualenv_root(self):
        stdin, stdout, stderr = self.run_cli()
        self.assertEqual(
            (stdin, stdout, stderr), ("", self.locator.root.path + "\n", ""),
        )

    def test_find_directory_with_binary(self):
        this_dir = FilePath(__file__).parent()
        stdin, stdout, stderr = self.run_cli(
            ["directory", this_dir.path, "python"],
        )
        this_dir_venv = self.locator.for_directory(this_dir)
        self.assertEqual(
            (stdin, stdout, stderr), (
                "",
                this_dir_venv.binary("python").path + "\n",
                "",
            ),
        )

    def test_find_existing_by_name_fails_for_non_existing_virtualenvs(self):
        stdin, stdout, stderr = self.run_cli(
            ["--existing-only", "name", "bla"], exit_status=1,
        )
        self.assertEqual((stdin, stdout, stderr), ("", "", ""))

    def test_find_existing_by_name_succeeds_for_existing_virtualenvs(self):
        path = MemoryPath(fs=self.fs, path=("bla",))
        path.createDirectory()
        stdin, stdout, stderr = self.run_cli(
            ["--existing-only", "name", "bla"],
        )
        self.assertEqual(
            (stdin, stdout, stderr), (
                "",
                self.locator.for_name("bla").path.path + "\n",
                "",
            ),
        )

    def test_find_existing_by_dir_fails_for_non_existing_virtualenvs(self):
        stdin, stdout, stderr = self.run_cli(
            ["--existing-only", "directory", "/foo/bla"], exit_status=1,
        )
        self.assertEqual((stdin, stdout, stderr), ("", "", ""))

    def test_find_existing_by_dir_succeeds_for_existing_virtualenvs(self):
        self.locator.root.child("bla").createDirectory()
        mem = MemoryPath(fs=self.fs, path=("foo", "bla",))
        stdin, stdout, stderr = self.run_cli(
            ["--existing-only", "directory", mem.path],
        )
        self.assertEqual(
            (stdin, stdout, stderr), (
                "",
                self.locator.for_directory(mem).path.path + "\n",
                "",
            ),
        )

    def test_cannot_specify_random_garbage(self):
        stdin, stdout, stderr = self.run_cli(
            ["--random-garbage"], exit_status=2,
        )
        self.assertEqual(
            (stdin, stdout, stderr),
            ("", stdout, "Error: no such option: --random-garbage\n"),
        )



from unittest import TestCase

from bp.memory import MemoryFS, MemoryPath

from mkenv.common import Locator, VirtualEnv


class TestLocator(TestCase):
    def test_named_virtualenvs_are_children(self):
        root = MemoryPath(fs=MemoryFS())
        locator = Locator(root=root)
        self.assertEqual(
            locator.for_name("one"),
            VirtualEnv(path=locator.root.child("one")),
        )
